/*
 * configfile.c
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      version 2 as published by the Free Software Foundation.
 *
 * Copyright (c) 2009 Thomas Bächler <thomas@archlinux.org>
 */

#ifndef CONFIGFILE
#error CONFIGFILE not defined
#endif
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "configfile.h"

/*
 * This configuration file parser parses everything in-place
 * While this should be efficient and safe, it is ugly and probably buggy
 */

static int ignore_line(const char *line) {
  /* Ignore empty lines and lines starting with #, also ignore spaces when checking */
  while(*line == ' ')
    line++;
  if(*line == '\0' || *line == '\n' || *line == '#')
      return 1;
  return 0;
}

static int pathname_match(const char *p1, const char *p2) {
  char *r2;
  int ret;

  /* Get the canonicalized path for p2, we assume this was already done for p1 */
  r2 = realpath(p2, NULL);
  if(r2 == NULL) {
    return 0;
  }
  ret = !strcmp(p1, r2);
  free(r2);
  return ret;
}

int getconfig(const char *path, struct pathcfg *cfg) {
  FILE *f = fopen(CONFIGFILE, "r");
  char *line = NULL, *opt, *opt2;
  size_t len;
  unsigned int nu, iu, ng, ig;
  int res = 0;

  while(getline(&line, &len, f) != -1) {
    if(line == NULL || ignore_line(line))
      continue;
    /* Get rid of the trailing newline */
    opt = index(line, '\n');
    if(opt != NULL)
      *opt = '\0';
    /* Get the path */
    opt = index(line, ' ');
    while(opt != line && opt != NULL && *(opt-1) == '\\') {
      memmove(opt-1, opt, strlen(opt)+1);
      opt = index(opt, ' ');
    }
    if(opt != NULL)
      *(opt++) = '\0';
    /* If the path matches the one from the commandline, continue */
    if((res = pathname_match(path, line)))
      break;
  }
  fclose(f);

  /* We didn't find the path in the configuration file */
  if(!res) {
    free(line);
    return 0;
  }

  /* Allocate 10 group and user char* pointers at once */
  nu = ng = 10;
  cfg->users = (char**)malloc(nu*sizeof(char*));
  cfg->groups = (char**)malloc(ng*sizeof(char*));
  iu = ig = 0;
  while(opt != NULL && *opt != '\0') {
    /* Skip whitespace */
    if(*opt == ' ') {
      opt++;
    }
    else {
      opt2 = index(opt, ' ');
      if(opt2 != NULL)
        *(opt2++) = '\0';

      if(opt[0] == '@') {
        /* We're using the last slot, add 10 new ones */
        if(ig == ng - 1)
          cfg->groups = realloc(cfg->groups, (ng+=10)*sizeof(char*));
        /* Copy the group name to the cfg structure */
        cfg->groups[ig] = malloc(strlen(++opt)+1);
        strcpy(cfg->groups[ig++], opt);
      }
      else {
        /* We're using the last slot, add 10 new ones */
        if(iu == nu - 1)
          cfg->users = realloc(cfg->users, (nu+=10)*sizeof(char*));
        /* Copy the user name to the cfg structure */
        cfg->users[iu] = malloc(strlen(opt)+1);
        strcpy(cfg->users[iu++], opt);
      }

      opt = opt2;
    }
  }
  /* Terminate with a NULL pointer */
  cfg->users[iu] = NULL;
  cfg->groups[ig] = NULL;

  free(line);
  return 1;
}

void getconfig_free(struct pathcfg *cfg) {
  char **c;

  /* Free all memory allocated in getconfig() */
  for(c = cfg->users; *c != NULL; ++c)
    free(*c);
  for(c = cfg->groups; *c != NULL; ++c)
    free(*c);

  free(cfg->users);
  free(cfg->groups);
}
